DAT programName byte "PropBoeSdDemo120428D", 0 ' This will let me know what program is loaded in the Prop.
CON
{{
  ******* Public Notes *******

  By: Duane Degn
  Date: April 28, 2012
  MIT license
  No rights reserved.
  
  The LEDs provide information about what the
  program is doing.

  When the SD card is being written to or read
  from the red LED will be on indicating the
  PropBOE should not be turned off.

  When the green LED is on, the PropBOE is safe
  to turn off.

  If there is a problem reading the SD cards near
  the beginning of the program neither the red
  or the green LED will turn on.


      
}}                     
{
  ******* Private Notes *******
  Modified from PropBoeBot120428A. 
  
}
CON

  _Clkmode = xtal1 + pll16x                           
  _Xinfreq = 5_000_000
   
  ' pin assignments

  _RedLedPin = 9                ' The red LED will be on while SD car is mounted.  Not safe (for log file)
                                ' to turn off PropBOE.
  _GreenLedPin = 8              ' SD card not currently mounted.
 
    
  _DoSdPin = 22                 ' SD card pin assignments.
  _ClkSdPin = 23
  _DiSdPin = 24
  _CsSdPin = 25
  _WpSdPin = -1                 ' Write Protect, not used (Micro SD cards don't have a write protect feature).
  _CdSdPin = -1                 ' Card Detect, not used for now.  I think this can be shared with DI pin.
  _RtcPin1 = -1                 ' Pins that would have been used by real time clock.
  _RtcPin2 = -1
  _RtcPin3 = -1
  

  _RefreshHz = 50
  _PstBaud = 57600
  _MaxFileNumber = 9999
  _ProgramVersionLocation = 17
  _VersionCharacterToUse = 3
  _VersionLocationInLogName = 1
  _LogNumberLocationInLogName = 4
  _BytesToBuffer = 11
  _RowsToRecord = 10

  _EndOfRecord = 9999

  '_CarriageReturn = 13

  ' sdFlag enumeration
  #0, _NotFoundSd, _InUseSd, _InitializingSd, _NewLogCreatedSd, _LogFoundSd, _NoLogYetSd, _InUseByOtherDeviceSd

  
  #0, _PlayBackSuccess, _PlayBackErrorFileNotFound, _PlayBackErrorOther

                                           
VAR

  long lastRefreshTime, refreshInterval
  long sdErrorNumber, sdErrorString, sdTryCount
  long numberToWrite[4]
  
  byte sdFlag
    
OBJ

  Pst : "Parallax Serial Terminal"
  Sd: "SD-MMC_FATEngine"
  
PUB Main 

  refreshInterval := clkfreq /  _RefreshHz

  dira[_RedLedPin] := 1
  outa[_RedLedPin] := 1
  dira[_GreenLedPin] := 1
  outa[_GreenLedPin] := 1
 
  sdFlag := _InitializingSd
  
  Decx(_EndOfRecord, 4, @endOfFile) 
        
  ' Change portion of log name to match program version.
  ' This will let me know which program created which file.   Commented out for now.
  'bytemove(@logName + _VersionLocationInLogName, @programName + _ProgramVersionLocation, _VersionCharacterToUse)
  
  Sd.fatEngineStart( _DoSdPin, _ClkSdPin, _DiSdPin, _CsSdPin, _WpSdPin, _CdSdPin, {
  '                  } _RtcPin1, _RtcPin2, _RtcPin3)
 
  Pst.Start(_PstBaud)
  
  waitcnt(clkfreq * 3 + cnt) ' time to open terminal window

  Pst.str(string(13, "Press any key to start."))

  Pst.charin

  Pst.str(string(13, 13, "PropBOE SD demo using ", 34, "SD-MMC_FATEngine", 34, " as the SD card driver.")) 
  Pst.str(string(13, "About to mount SD card.")) 
  Pst.str(string(13, "If no new text appears, there was a problem mounting the SD card.")) 
  Pst.str(string(13, "If the SD card fails to mount try re-seating the card in the")) 
  Pst.str(string(13, "socket (after turning off power to PropBOE).")) 
  Pst.str(string(13, "If that doesn't work, open the SD card on a PC."))
  Pst.str(string(13, "There might be a ", 34, "check disk", 34, " flag that needs to be cleared.")) 
  Pst.str(string(13, "The final thing to try if the SD card still wont mount is to reformat it."))
  Pst.str(string(13, "I believe the cluster size should be 32MB?.")) 
  Pst.str(string(13, "I've always just used the default settings when reformatting.")) 
  Pst.str(string(13, "The ", 34, "quick", 34, " reformatting method hasn't ever worked for me.")) 
  Pst.str(string(13, "It needs to be the standard (long way) of reformatting.")) 
  Pst.str(string(13, "I've had several SD cards that would not work with the above")) 
  Pst.str(string(13, "mentioned SD driver until the card had been reformatted.")) 
  Pst.str(string(13, 13, "****** Attempting to Mount SD Card ******", 13)) 
  
  MountSd


  ' Open log files on SD card to see how many there are.  (There is likely be a better way of doing this.)
  repeat while sdFlag == _InitializingSd and sdTryCount < _MaxFileNumber
    
    Pst.str(string(13, "Looking for file ", 34))
    Pst.str(@logName)
    Pst.char(34)
    sdErrorString := \Sd.openFile(@logName, "R")
    if strcomp(sdErrorString, @logName)
      Pst.str(string(13, "File ", 34))
      Pst.str(@logName)
      Pst.str(string(34, " found."))
      Pst.str(string(13, "Closing File"))
      Sd.closeFile
      Decx(++sdTryCount, 4, @logName + _LogNumberLocationInLogName)  
    else
      Pst.str(string(13, "File ", 34))
      Pst.str(@logName)
      Pst.str(string(34, " not found."))
      Pst.str(string(13, "sdErrorString = "))
      Pst.dec(sdErrorString)
      Pst.str(string(13, "sdErrorString ", 34))
      Pst.str(sdErrorString)
      Pst.char(34)
      if sdTryCount == 0
        sdFlag := _NoLogYetSd
        Pst.str(string(13, "There are not presently any log files on this SD card."))
      else
        sdFlag := _LogFoundSd
        Pst.str(string(13, "There are presently "))
        Pst.dec(sdTryCount)
        Pst.str(string(" log files on this SD card.")) 

  ' sdTryCount will one unit larger than the file number.
  ' if sdTryCount is zero then there aren't any files yet.
  Pst.str(string(13, "Unmounting Partition"))
  sdErrorNumber := Sd.unmountPartition
  Pst.str(string(13, "sdErrorNumber = "))
  Pst.dec(sdErrorNumber)
  outa[_GreenLedPin]~~
  outa[_RedLedPin]~

  StartLog

  Pst.str(string(13, "About to write "))
  Pst.dec(_RowsToRecord)
  Pst.str(string(" sets (rows) of 4 data items (columns) to uSD."))
    
  WriteData

  EndRecording

  Pst.str(string(13, "The data has now been written to the uSD."))

  Pst.str(string(13, "Preparing to retrieve data from the uSD."))

  OpenFileToRead(sdTryCount - 1) ' "sdTryCount" is one larger than the file number just written.

  DisplayData

  EndReading

  Pst.str(string(13, "End of uSD file demo."))

  repeat
    Pst.str(string(13, "Press ", 34, "r", 34, " to reboot."))
    result := Pst.charin
    if result == "r"
      reboot
      
PUB WriteData | localIndex, localSize

  lastRefreshTime := cnt '+ refreshInterval
  numberToWrite[0] := 1000    ' dummy data
  numberToWrite[1] := 2000
  numberToWrite[2] := 3000
  numberToWrite[3] := 4000
    
  repeat _RowsToRecord
    numberToWrite[0] += 1   ' each dummy data is changed by a different amount.
    bytefill(@sdBuffer, 0, _BytesToBuffer) ' Clear buffer
    
    ' The SD object doesn't have a way to write numbers as ASCII characters to the file
    ' so we need to convert the number to an ASCII string before writing them to the file.
    ' The method "Dec" (and friends) converts a number to ASCII character and stores them
    ' in hub RAM.
    Dec(numberToWrite[0], @sdBuffer)
    localSize := strsize(@sdBuffer)
    Sd.writeData(@sdBuffer, localSize) ' write first number of set to SD card
    Sd.writeByte(",")                     ' write data separater "," to SD card
    
    numberToWrite[1] += 10
    bytefill(@sdBuffer, 0, _BytesToBuffer) ' Clear buffer 
    Dec(numberToWrite[1], @sdBuffer)
    localSize := strsize(@sdBuffer)
    Sd.writeData(@sdBuffer, localSize)
    Sd.writeByte(",")                    
    
    numberToWrite[2] += 100
    bytefill(@sdBuffer, 0, _BytesToBuffer) ' Clear buffer 
    Dec(numberToWrite[2], @sdBuffer)
    localSize := strsize(@sdBuffer)
    Sd.writeData(@sdBuffer, localSize)
    Sd.writeByte(",")                    
    
    numberToWrite[3] += 1000
    bytefill(@sdBuffer, 0, _BytesToBuffer) ' Clear buffer 
    Dec(numberToWrite[3], @sdBuffer)
    localSize := strsize(@sdBuffer)
    Sd.writeData(@sdBuffer, localSize)                 
    
    ' Add end of line characters so file will display in Excel correctly.
    Sd.writeByte(13)            
    Sd.writeByte(10)
    
PUB DisplayData | localSize
'' This method just dumps the file contents to the terminal window.
'' See my other PropBOE-Bot code for an example of using the data
'' from the file in a more meaningful way.

  localSize := Sd.fileSize
  Pst.char(13) 
  Pst.char(13) 
  repeat localSize
    Pst.char(Sd.readByte)
        
PUB StartLog | logNumberToWrite
'' Open log file on SD card.

  outa[_GreenLedPin]~
  outa[_RedLedPin]~~
  MountSd 
  if sdFlag == _NoLogYetSd
    logNumberToWrite := 0
  else
    logNumberToWrite := sdTryCount
    if sdFlag == _NewLogCreatedSd
      sdFlag := _LogFoundSd
  Decx(logNumberToWrite, 4, @logName + _LogNumberLocationInLogName) 
  repeat while (sdFlag <> _NotFoundSd and sdFlag <> _NewLogCreatedSd) and sdTryCount < _MaxFileNumber
    Pst.str(string(13, "Atempting to create file ", 34))
    Pst.str(@logName)
    Pst.char(34)
    sdErrorString := \Sd.newFile(@logName)
    sdErrorNumber :=  Sd.partitionError ' Returns zero if no error occurred.
     
    if(sdErrorNumber) ' Try to handle the "entry_already_exist" error.
      if(sdErrorNumber == Sd#Entry_Already_Exist)
        ' This section of code does not change "sdFlag" so this
        ' loop will repeat and try to open the next file number.
        
        'repeat until not lockset(debugLockID) 
        Pst.str(string(13, "File ", 34))
        Pst.str(@logName)
        Pst.str(string(34, " already exists. (This shouldn't happen!)"))
        'lockclr(debugLockID)
        Decx(++sdTryCount, 4, @logName + _LogNumberLocationInLogName)
        
     
      else
        'repeat until not lockset(debugLockID) 
        Pst.str(string(13, "Create File Errror = "))
        Pst.dec(sdErrorNumber)
        'lockclr(debugLockID)
        outa[_GreenLedPin]~
        outa[_RedLedPin]~~
        repeat 200 ' Flash red LED to let user know something is wrong.
          !outa[_RedLedPin]
          waitcnt(clkfreq / 10 + cnt)
        sdFlag := _NotFoundSd
        outa[_GreenLedPin]~~  ' Both LEDs off to show indicate SD problem.
        outa[_RedLedPin]~~
    else
      sdFlag := _NewLogCreatedSd
      'repeat until not lockset(debugLockID) 
      Pst.str(string(13, "Creating file ", 34))
      Pst.str(@logName)
      Pst.str(string(34, "."))     
      'waitcnt(clkfreq / 4 + cnt)  
      Sd.openFile(@logName, "W")
      sdTryCount++
       
PRI EndRecording

  
  Sd.writeString(@endOfFile)   ' An end of file indicator.  Not used in this program.
  Sd.writeByte(13)
  Sd.writeByte(10)
  Sd.writeString(string("Program used to create log "))
  Sd.writeString(@programName)  ' I like to know what program generated the log file.
  Sd.writeByte(13)
  Sd.writeByte(10)
  
  Pst.str(string(13, "EndRecording Method"))
  
  UnmountSd
  outa[_GreenLedPin]~~    ' Okay to remove SD card.
  outa[_RedLedPin]~
    
PRI EndReading

  UnmountSd
  outa[_GreenLedPin]~~
  outa[_RedLedPin]~

PRI OpenFileToRead(fileToPlay)

  if fileToPlay < 0
    Pst.str(string(13, "Error in program.  Stopped at BeginPlayBack method."))
    result := _PlayBackErrorOther
    outa[_GreenLedPin]~~
    outa[_RedLedPin]~~
    return
    
  result := MountSd
  if result == 1
    result := _PlayBackErrorOther
    outa[_GreenLedPin]~~
    outa[_RedLedPin]~~
    return
  outa[_GreenLedPin]~
  outa[_RedLedPin]~~
  Decx(fileToPlay, 4, @logName + _LogNumberLocationInLogName)
  Pst.str(string(13, "Looking for file ", 34))
  Pst.str(@logName)
  Pst.char(34)
  sdErrorString := \Sd.openFile(@logName, "R")
  if strcomp(sdErrorString, @logName)
    Pst.str(string(13, "File ", 34))
    Pst.str(@logName)
    Pst.str(string(34, " found."))
    result := _PlayBackSuccess
  else
    Pst.str(string(13, "File ", 34))
    Pst.str(@logName)
    Pst.str(string(34, " not found."))
    Pst.str(string(13, "sdErrorString = "))
    Pst.dec(sdErrorString)
    Pst.str(string(13, "sdErrorString ", 34))
    Pst.str(sdErrorString)
    Pst.char(34)
    UnmountSd
    outa[_GreenLedPin]~~
    outa[_RedLedPin]~~
    'sdFlag := _NotFoundSd
    result := _PlayBackErrorFileNotFound

PUB Dec(value, localPtr)
'' Print a decimal number
  result := decl(value, 10, 0, localPtr)
  'byte[result] := _CarriageReturn
  
PUB Decf(value, width, localPtr)
'' Prints signed decimal value in space-padded, fixed-width field
  result := decl(value, width, 1, localPtr)

PUB Decx(value, digits, localPtr)
'' Prints zero-padded, signed-decimal string
'' -- if value is negative, field width is digits+1
  result := decl(value, digits, 2, localPtr)

PUB Decl(value, digits, flag, localPtr) | localI, localX
'' DWD Fixed with FDX 1.2 code

  digits := 1 #> digits <# 10
  
  localX := value == NEGX       'Check for max negative
  if value < 0
    value := ||(value + localX) 'If negative, make positive; adjust for max negative
    byte[localPtr++] := "-"
  localI := 1_000_000_000
  if flag & 3
    if digits < 10                                      ' less than 10 digits?
      repeat (10 - digits)                              '   yes, adjust divisor
        localI /= 10

  repeat digits
    if value => localI
      byte[localPtr++] := value / localI + "0" + localX * (localI == 1)
      value //= localI
      result~~
    elseif (localI == 1) OR result OR (flag & 2)
      byte[localPtr++] := "0"
    elseif flag & 1
      byte[localPtr++] := " "
    localI /= 10
  result := localPtr
    
PUB MountSd

  sdErrorNumber := Sd.mountPartition(0)
  if sdErrorNumber
    sdFlag := _NotFoundSd
    result := 1
    Pst.str(string(13, "Error Mounting SD Card #"))
    Pst.dec(sdErrorNumber)
    outa[_GreenLedPin]~
    outa[_RedLedPin]~~
    repeat 25   ' Flash LEDs to show there was a problem.
      !outa[_RedLedPin]
      !outa[_GreenLedPin]
      waitcnt(clkfreq / 4 + cnt)
    outa[_GreenLedPin]~  ' Since I'm using red to indicated the SD card is active
                         ' and the PropBOE shouldn't be turned off, I need to
                         ' Leave both LEDs off indicate SD failure.                               
    outa[_RedLedPin]~
    if sdErrorNumber == -1
      Pst.str(string(13, "The SD card was not properly unmounted after its last use."))
      Pst.str(string(13, "Continuing with program."))
      sdFlag := _InitializingSd
      outa[_GreenLedPin]~ 
      outa[_RedLedPin]~~
      result := 0
  else
    outa[_GreenLedPin]~ 
    outa[_RedLedPin]~~   ' When the red LED is on the PropBOE shouldn't be turned off.

PUB UnmountSd

  Pst.str(string(13, "Unmounting Partition"))
  sdErrorNumber := Sd.unmountPartition
  Pst.str(string(13, "sdErrorNumber = "))
  Pst.dec(sdErrorNumber)
  outa[_GreenLedPin]~~
  outa[_RedLedPin]~

DAT

logName       byte "LOG_0000.TXT", 0  ' Use all caps in file names or SD driver wont find them.

sdBuffer      byte 0[_BytesToBuffer + 1]
endOfFile     byte 0[5]  ' this will be filled by program
    